/*
 * Test pipe operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h> /* sleep */

#include <fsclient/client.h>
#include <mem/memory_utils.h>



/* Minimum pipe region size in pages. */

const unsigned int PIPE_PAGES = 2;

/* Use the writer to fill the pipe with data. */

static void write_pipe(file_t *writer, int number)
{
  offset_t size = 600;
  char buffer[size];
  long err;

  /* Make writers blocking to permit synchronisation. */

  if ((err = client_set_blocking(writer, NOTIFY_SPACE_AVAILABLE)))
  {
    printf("Could not set pipe #%d as blocking: %s\n", number, l4sys_errtostr(err));
    return;
  }

  for (int loop = 0; loop < 3; loop++)
  {
    for (int region = 0; region < 26; region++)
    {
      memset(buffer, (int) 'a' + region, size);

      offset_t nwritten = client_write(writer, buffer, size);

      printf("Written %ld/%ld in #%d of %d/%d to pipe #%d...\n", nwritten, size, region, loop, 2, number);
    }

    sleep(1);
  }

  /* Flush to make the final output available. */

  client_flush(writer);

  client_close(writer);
}

/* Use the reader to obtain data from the pipe. */

static void read_pipes(file_t *reader1, file_t *reader2)
{
  offset_t size = 600, totals[] = {0, 0};
  unsigned int active = 2;
  long err;
  file_t *reader;

  /* Register the readers for notification. */

  // NOTE: Use the flags to detect initial conditions!

  if ((err = client_subscribe(reader1, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED)) ||
      (err = client_subscribe(reader2, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED)))
  {
    printf("Could not subscribe to notifications: %s\n", l4sys_errtostr(err));
    return;
  }

  while (1)
  {
    char buffer[size];
    offset_t nread;

    /* Wait for notification of content. */

    err = client_wait_files(&reader);

    if (err)
    {
      printf("Error waiting for notifications: %s\n", l4sys_errtostr(err));
      return;
    }

    if ((reader != reader1) && (reader != reader2))
    {
      printf("Spurious notification received for %p versus %p, %p.\n", reader, reader1, reader2);
      continue;
    }

    int p = reader == reader1 ? 0 : 1;

    printf("Pipe #%d notified with conditions:%s%s\n", p, reader->notifications & NOTIFY_PEER_CLOSED ? " closed" : "",
                                                          reader->notifications & NOTIFY_CONTENT_AVAILABLE ? " content" : "");

    nread = client_read(reader, buffer, size);

    while (nread)
    {
      totals[p] += nread;

      printf("Read %ld/%ld, total %ld, first %c, last %c, from pipe #%d...\n", nread, size, totals[p], *buffer, *(buffer + nread - 1), p + 1);
#if 0
      for (offset_t i = 0; i < nread; i += 60)
      {
        fwrite(buffer + i, sizeof(char), nread - i > 60 ? 60 : nread - i, stdout);
        fputs("\n", stdout);
      }
#endif
      nread = client_read(reader, buffer, size);
    }

    /* Without any more content, a peer closed event should terminate reading
       from the pipe. */

    if (reader->notifications & NOTIFY_PEER_CLOSED)
    {
      active--;
      if (!active)
        break;
    }
  }

  client_close(reader1);
  client_close(reader2);

  printf("Data shown.\n");
}

int main(void)
{
  /* Invoke the open method to receive the pipe endpoint references. */

  file_t *reader1, *reader2, *writer1, *writer2;
  long err = client_pipe(&reader1, &writer1) || client_pipe(&reader2, &writer2);

  if (err)
  {
    printf("Could not obtain pipes: %s\n", l4sys_errtostr(err));
    return 1;
  }

  /* Schedule reader and writer threads. */

  std::thread *activities[3];

  activities[0] = new std::thread(read_pipes, reader1, reader2);
  activities[1] = new std::thread(write_pipe, writer1, 1);
  activities[2] = new std::thread(write_pipe, writer2, 2);

  for (int i = 0; i < 3; i++)
    activities[i]->join();
}

// vim: tabstop=2 expandtab shiftwidth=2
